/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package friedrich
import scala.xml.XML
import friedrich.util._

/**
 * Configuration data for a Friedrich application.
 */
trait Configuration {
	var verboseMemory: Boolean = false
	var gcBetweenPhases: Boolean = true
	var pipeline: String = "default"
	var pipelines: Map[String, Seq[String]] = Map()
}

class BasicConfiguration extends Configuration 
  
/**
 * This class is responsible for resolving the configuration options
 * for a given configuration object.
 * In order to use this class, at a bare minimum, blankConfiguration should be
 * implemented and populateConfiguration should be overridden.
 * 
 * In general, options are first read from the command line, then from an XML
 * file, and finally from defaults, if any.
 */
abstract class ConfigReader[T <: Configuration](args: Seq[String]) extends CmdLineOptions {
  
  val configName = stringOption(args, "-config").getOrElse("default")
  val xmlOdict = readConfiguration()
  
   /**
   * Read the default configuration from an XML file.
   */
  private def readConfiguration(): Map[String, String] = {
    println("Reading configuration " + configName)
    readConfiguration(configName)
  }
 
  private def fileExists(f: String) = new java.io.File(f).exists
  private def missingConfigurationMessage() = stringOption(args, "-config") match {
    case None => {}
	case Some(config) => { println("Requested configuration " + config + " not found") }
  }
  
  /**
   * Attempt to read the named configuration from config.xml.
   */
  protected def readConfiguration(name: String): Map[String, String] = {
		if (!fileExists("config.xml")) {
			missingConfigurationMessage()
            Map() //not found
		} else {
			val f = XML.loadFile("config.xml")
			val configurations = f \\ "configuration"

			val myConfig = configurations.find(_(0).attributes("name").text == name)
			if (myConfig == None) {
				missingConfigurationMessage()
				Map() //not found
			} else {
				val options = myConfig.get \\ "option"
				Map() ++ options.map(x => { ((x.attributes("name").text, x.text)) } )            
			}
		}
  }
 
  /**
   * Read pipelines from config.xml.
   */
  def readPipelines():Map[String, Seq[String]] = {
	    val f = XML.loadFile("config.xml")
	    val pipelines = f \\ "pipeline" map (pl => {
	        val name = pl \ "@name" text
	        val phases = pl \ "phase" map (_.text)
	        (name -> phases)
	    })
	    pipelines.toMap
  }
 
  /**
   * Subclasses should implement this method and create
   * a new blank configuration object.
   */
  def blankConfiguration(): T
  
  /**
   * Extending classes should implement this method to 
   * fill in the configuration object. Methods such as resolveString
   * and resolveInt can be used to fill in the values of the config object
   * properly. 
   * Overriding methods should call the superclass method first. 
   */  
  def populateConfiguration(config: T) = {
    config.pipelines = readPipelines()
    config.pipeline = resolveString("-pipeline", "pipeline", "default")
    config.verboseMemory = resolveBoolean("-verboseMemory", "verboseMemory")
    config.gcBetweenPhases = resolveBoolean("-gcBetweenPhases", "gcBetweenPhases")
  }
  
  /**
   * Obtain a populated configuration.
   */
  def configuration(): Option[T] = {
    val r = blankConfiguration()
    populateConfiguration(r)
    Some(r)
  }
  
  def resolveBoolean(flag: String, key: String): Boolean = {    
	booleanOption(args, flag) || xmlOdict.get(key).map(_ == "true").getOrElse(false)
  }
  
  /**
   * Resolve a string option by consulting first the command line, then the 
   * XML config file.
   */
  def resolveString(flag: String, key: String): Option[String] = {
	stringOption(args, flag).orElse(xmlOdict.get(key))
  }
  
  /**
   * Resolve a string option by consulting first the command line, then the 
   * XML config file, and finally the default value.
   */
  def resolveString(flag: String, key: String, default: String): String = {
	stringOption(args, flag).orElse(xmlOdict.get(key)).getOrElse(default)
  }
  
     /**
   * Resolve an int option by consulting first the command line, then the 
   * XML config file.
   */
  def resolveInt(flag: String, key: String): Option[Int] = {
	stringOption(args, flag).orElse(xmlOdict.get(key)).map(_.toInt)
  }
  
  /**
   * Resolve an int option by consulting first the command line, then the 
   * XML config file, and finally the default value.
   */
  def resolveInt(flag: String, key: String, default: Int): Int = {
	stringOption(args, flag).orElse(xmlOdict.get(key)).map(_.toInt).getOrElse(default)
  }
 
  /**
   * If this method wraps a call to resolveInt or resolveString (the varieties
   * that return an option), an error will result when a necessary option is missing.
   */
  protected def require[T](parameter: Option[T], message: String): T = parameter match {
    case None => {
      Console.err.println(message)
      usage()
      sys.exit(1)
    }
    case Some(x) => x
  }
   
  def usage() {
    
  }
}

/**
 * Bare-bones configuration reader with empty implementations.
 */
class BasicConfigReader(args: Seq[String]) extends ConfigReader[BasicConfiguration](args) {
  def blankConfiguration() = new BasicConfiguration()
}