/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package friedrich

/**
 * Pipeline management functions.
 * Todo:
 * "Stop after"/"resume" functionality.
 */
object Pipeline {
  /**
   * Construct a pipeline object from the given phase names (class names).
   * This uses reflection to instantiate the classes.
   */
  def fromPhaseNames[T <: PhaseData](phaseNames: Seq[String]) : Pipeline[T] = {
      var phases = List[Phase[T]]()
       
        //load the phases
        for (p <- phaseNames) {
            val c = Class.forName(p)
                val cons = c.getConstructor() //no-arg constructor must exist!
                val phase = cons.newInstance().asInstanceOf[Phase[T]]
                println("Loaded phase " + phase)
                phases ::= phase
        }       
       new Pipeline(phases.reverse)
  }
  
}

/**
 * Construct a pipeline from the given phase objects.
 */
class Pipeline[T <:PhaseData](phases: Seq[Phase[T]]) {
	var timeTaken = Map[String, Double]()
	
	/**
	 * Run the phases in this pipeline.
	 */
	def runPhases(data: T) {
       for (p <- phases) {
    	  val start = System.currentTimeMillis()
          p.run(data)
          val end = System.currentTimeMillis()
          val taken = (end - start)/1000.0
          println("Time taken for this phase: " + taken + " s")
          timeTaken += ((p.toString, taken))
       }
       timeBreakdown()
    }
	
	/**
	 * Display a breakdown of the time taken for the various phases.
	 */
	def timeBreakdown() = {
	  for ((k, v) <- timeTaken) {
	    printf("%s: %.2f s\n", k, v)
	  }
	}
    
}
