/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package friedrich.graph

import miniasm.genome._
import friedrich.graph._

import scala.collection.immutable.HashMap
import scala.collection.immutable.TreeSet
import scala.collection.mutable.{Set, MultiMap}
import scala.collection._

/**
 * A mutable graph that represents data internally as an adjacency list.
 * Adjacency lists map each node to a list of adjacent nodes.
 * 
 * In the case of genome assembly, the graphs are quite sparse. Many
 * nodes only have two adjacent nodes, "previous" and "next". In order
 * to save memory and time in this common case, we store this information
 * in the graph node itself (prevNode and nextNode) unless there are 
 * more than two adjacent nodes. If this is the case, then the adjacency list
 * is "migrated" to this graph object and stored here instead, in
 * adjList and revAdjList.
 * 
 * TODO: override hasEdge 
 */
trait AdjListGraph[N] extends Graph[N] {

	protected val adjList = new mutable.HashMap[N, Set[N]] with MultiMap[N, N]		
	protected val revAdjList = new mutable.HashMap[N, Set[N]] with MultiMap[N, N]
	
	final override def edges(from: N): Iterable[N] = edgesFrom(from) ++ edgesTo(from)
	
	final def edges(): Iterable[(N, N)] = nodes.flatMap( (n:N) => {
	  edgesFrom(n).map((x:N) => (n,x))
	})
	
	override def numEdges = nodes.foldLeft(0)(_ + edgesFrom(_).size)
	  
	def edgesFrom(from: N): Iterable[N] = adjList.getOrElse(from, Set())
	def edgesTo(to: N): Iterable[N] = revAdjList.getOrElse(to, Set())
	
	def implAddEdge(from: N, to: N) {
	  adjList.addBinding(from, to)
	  revAdjList.addBinding(to, from)
	}

}
