/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package friedrich.graph

import scala.collection.immutable.{HashMap, HashSet}
import scala.collection.mutable

/**
 * Base trait for graph implementations. 
 * A graph is a set of nodes and edges.
 */
trait Graph[N] {
	
	def nodes: Iterable[N]
	
	def numEdges: Int = edges.size
	def numNodes: Int = nodes.size
	
	def edgesFrom(from: N): Iterable[N]
	def edgesTo(to: N): Iterable[N]
	/**
	 * Both to- and from-nodes of the given node
	 */
	def edges(node: N): Iterable[N] = edgesFrom(node) ++ edgesTo(node)
	/**
	 * All edges in the graph
	 */
	def edges(): Iterable[(N, N)] 
	
	def hasEdge(from: N, to: N): Boolean = edgesFrom(from).iterator.contains(to)	
	def hasNode(node: N): Boolean = nodes.iterator.contains(node)
	
	/**
	 * Users should call this method to add a node.
	 */
	def addNode(node: N): Unit = {
	  if (!hasNode(node)) {
	    implAddNode(node)
	  }
	}
	
	/**
	 * Users should call this method to add an edge.
	 */
	def addEdge(from: N, to: N): Unit = {
		assert(hasNode(from))
		assert(hasNode(to))
		if (!hasEdge(from, to)) {						
			implAddEdge(from, to)			
		}
	}
	
	/**
	 * This method should be overridden to implement the operation of adding a node.
	 */
	protected def implAddNode(node: N)		
	
	/**
	 * This method should be overridden to implement the operation of adding an edge.
	 */
	protected def implAddEdge(from: N, to: N) 
	
	/**
	 * Print a textual representation of the graph.
	 * For debug purposes.
	 */
	def printBare() = {
		for (n <- nodes) {
			println(n)
			print(n + " to: ")
			for (e <- edgesFrom(n)) {				
				print(e + ", ")
			}
			print(" from: ")
			for (e <- edgesTo(n)) {				
				print(e + ", ")
			}
			println("")
		}
	}
	
}

/**
 * Trait for a graph that stores nodes in a mutable set.
 */
trait SetBackedGraph[N] extends Graph[N] {
  	final def implAddNode(node: N) {
		nodesv.add(node)
	}
  		
	final def nodes(): Iterable[N] = nodesv
	
	private val nodesv = new mutable.HashSet[N]
	
	final override def hasNode(node: N) = nodesv.contains(node)		
}

