/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package miniasm
import miniasm.util._
import friedrich._
import miniasm.genome._
import miniasm.genome.bpbuffer.BPKmer
import scala.collection.immutable.{HashSet, HashMap}
import java.io.FileWriter
import miniasm.genome.bpbuffer.BPKmerFactory

/**
 * The main assembler application.
 * In order to change the type of Kmer being used, it is necessary to change
 * Assembler.T and the data.factory assignment in initData().
 */
object Assembler extends FriedrichApp[AssemblyData[BPKmer]] {
    implicit def toShort(i: Int) = i.asInstanceOf[Short]
      
    type T = miniasm.genome.bpbuffer.BPKmer                                                      
    
    def newData(): AssemblyData[T] = new AssemblyData[T](ClassManifest.fromClass(classOf[T]))
    
    /**
     * Initialise the configuration and an empty data object.
     * This method is also suitable for interactive use, to get a starting point
     * for working with the phases.
     */
    override def initData(args: Seq[String]): AssemblyData[T] = {
        val data = newData()      
        data.factory = new BPKmerFactory()
        val cr = new miniasm.util.ConfigReader(args)
        var config = cr.configuration()
        
        if (config == None) {    
        	println("No suitable configuration found in config.xml.")
        	sys.exit(1)            
        }                   
        data.config = config.get        
        data
    }
  
    def main(args: Array[String]) {       
        val data = initData(args)
        val config = data.config        
        val pipeline = config.pipeline
        println("Pipeline: " + pipeline)
        
        println("Assembly starting. Parameters: ")
        println("\tk %d \tminLen %d \tinput %s \tcutoff %s".format(data.k, config.minKmers, config.input, config.cutoff))
        
        runPipeline(data, pipeline)
        
        println("Printing contigs")            
        println(data.paths.size + " paths")
        writeContigFile(data, config.minKmers)
        println("Done")       
    }
        
    def writeContigFile(data: AssemblyData[T], minLen: Int) {
        var count = 1
        val outFile = new FileWriter("contigs.fa")
        val statsFile = new FileWriter("contigs.info")
        var seq = ""

        for (p <- data.paths; if p._1.size >= minLen) {     
            //note that p._1.size is number of kmers	            
        	val contigBps = p._1.size + data.k - 1
            outFile.write(">contig" + count + "_" + contigBps + "bp\n")
            println(contigBps)
            //println(p)

            for (node <- p._1) {
                node.contig = count
            }

            seq = FindPathsPhase.pathToString(p._1, data.k)
            outFile.write(seq + "\n")
            println(seq)
            
            statsFile.write(">contig" + count + "_" + contigBps + "bp\t" + p._2 + "\t" + p._3 + "\n")

            count += 1
        }

        outFile.close()        
        statsFile.close()
    }
    
}
