/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm

import miniasm.util._
import miniasm.genome._
import friedrich._
import friedrich.graph._

object BuildGraphPhase {
  /** 
   * Convenience function for interactive use
   */
  def apply[T <: Kmer[T]](data: AssemblyData[T]) = (new BuildGraphPhase[T]).run(data)
}

/**
 * Phase to build a graph from reads and kmers.
 * Input: data.reads, data.k
 * Output: data.graph
 */
class BuildGraphPhase[T <: Kmer[T]] extends AsmPhase[T] with DNAHelpers {
	
	
	def runImpl(data: AssemblyData[T]) {
		val g = new KmerSetBackedPathGraph[T](data.k, data.kmers)(data.nodeManifest)
	
		var count = 0
		
		print("Nodes")
		for (read <- data.reads; kmer <- read.kmers) {
			g.addNode(kmer)
			count += 1
			if (count % 100000 == 0) {
				print (".")
			}
			if (count % 10000000 == 0) {
				println("")
			}
		}
		println("")
		
		count = 0
		
		print("Edges")
		for (read <- data.reads) {
			val kmers = read.kmers
			
			for (i <- 0 to kmers.size - 2) {				
				//insert edges that were present together in a read
				g.addEdge(kmers(i), kmers(i+1))
				
				count += 1
				if (count % 100000 == 0) {
					print (".")
				}
				if (count % 10000000 == 0) {
					println("")
				}
			}
			
			insertStartFinish(g, data.kmers, kmers.head, kmers.last, data.k)
		}
		println("")
				
		println(g.nodes.size + " nodes")				
		println(g.numEdges + " edges")
//		println(g.migrations + " migrations")
		
//		g.printBare()
				
		data.graph = g		
	}
	
	//look for places to join in the first and last kmer of a read into the graph
	def insertStartFinish(g: Graph[T], kmerSet: KmerSet[T], start: T, finish: T, k: Int)  = {
		
		//Look for possible kmers that continue from "finish" and that are in the same kmer set.
		//They are guaranteed to already be nodes in the graph.
		val cns = finish.continuationsInSet
		
		for (c <- cns) {
			g.addEdge(finish, c)			
//			println(finish + " -> " + c)		
		}
		
		//Look for possible kmers that precede "start" and that are in the same kmer set.
		//They are guaranteed to already be nodes in the graph.
		val invCns = start.continuesFromInSet
		
		for (c <- invCns) {
			g.addEdge(c, start)				
//			println(c + " -> " + start)			
		}
		
	}
	
}
