/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm

import miniasm.util._
import miniasm.genome.Kmer

import scala.collection.mutable.HashMap
import java.io.FileWriter
import friedrich.graph._

object ComputeCoveragePhase {
  /** 
   * Convenience function for interactive use
   */
  def apply[T <: Kmer[T]](data: AssemblyData[T]) = (new ComputeCoveragePhase[T]).run(data)
}

/**
 * This phase inspects kmers and sets their "noise" flag if they have
 * low coverage. It also computes a coverage histogram.
 */
class ComputeCoveragePhase[T <: Kmer[T]]() extends AsmPhase[T] {

    def runImpl(data: AssemblyData[T]): Unit = { 
        var thresh = 0
        var cov_stats = computeHistogram(data.graph)
        val cutoff = data.config.cutoff
        if (cutoff == None) {
            thresh = findMinimum(cov_stats._3)
        }
        else {
            thresh = cutoff.get
        }
        ignoreLowCov(data.graph, thresh)
    }

    def computeHistogram(g: Graph[T]): Tuple3[Float, Int, HashMap[Int, Int]] = {
        // computes and prints to file a coverage histogram
        // returns a 3-tuple with mean and median coverage, and the histogram
        var covmap = new HashMap[Int, Int]
        var max = 0
        var totalcov = 0

        for (n <- g.nodes) {
            totalcov += n.coverage
            if (n.coverage > max)  { max = n.coverage }
            if (covmap.contains(n.coverage)) {
                covmap(n.coverage) += 1
            }
            else {
                covmap += n.coverage -> 1
            }
        }
        
        val avcov = totalcov/g.nodes.size
        println("Average coverage: " + avcov)

        val outfile = new FileWriter("coverage.hist")

        var count = 0
        val thresh = g.nodes.size/2
        var median = 0
        var found = false

        for (i <- 1 to max; if covmap.contains(i)) {
            outfile.write(i + "\t" + covmap(i) + "\n")
            count += covmap(i)
            if ((count > thresh) && (!found)) { median = i; found = true }
        }

        outfile.close()

        Tuple3(avcov, median, covmap)
    }

    def findMinimum(covmap: HashMap[Int, Int]): Int = {
        // finds the first local minimum in the coverage histogram
        println("Finding minimum")
        var current = covmap(1)
        var i = 2
        var found = false
        while ((!found) && (i < covmap.keys.max)) {
            if (covmap.contains(i)) {
                if (covmap(i) > current) { found = true; } else { current = covmap(i); i += 1 }
            }
            else {
                i += 1
            }
        }
        if (found == false)
            i = 0
        i
    }

    def ignoreLowCov(g: Graph[T], min: Int): Unit = {
        // ignores kmers with coverage lower than min
        println("Ignoring nodes with coverage < " + min + "...")
        var count = 0
        for (n <- g.nodes) {
          if (n.coverage < min) {
            n.noise = true
            count += 1
          } else {
            //reset the flag in this case, so that if this function is run multiple times
            //with different thresholds, the  old results are not preserved
            n.noise = false   
          }
        }
        println(count + " nodes ignored!")
    }
}
