/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm

import miniasm.util._
import miniasm.genome._
import scala.collection.mutable.{HashSet, ArrayBuffer}
import scala.annotation.tailrec
import friedrich.graph._
import friedrich.graph._

object FindPathsPhase {
  /** 
   * Convenience function for interactive use
   */
	def apply[T <: Kmer[T]](data: AssemblyData[T]) = (new FindPathsPhase[T]).run(data)
  
	def printPath[T <: Kmer[T]](p: Iterable[T], k: Int) {
		println(p.size)
//			println(p)
        println("DEBUG: " + p.head)
        if (p.size > 1)
            println(pathToString(p, k))
	}
	
	def pathToString[T <: Kmer[T]](p: Iterable[T], k: Int): String = {
		if (p.size > 1) {
			p.head.toString + p.drop(1).map {_.toString.charAt(k - 1) + ""}.reduceLeft(_+_)
		} else if (p.size == 1) {
		  p.head.toString
		} else {
			""
		}
	}		
}

/**
 * Phase to extract continuous paths from a graph.
 * Input: the kmer set in 'data' has been configured and populated.
 * Output: contigSet in 'data' will be configured and populated.
 */
class FindPathsPhase[T <: Kmer[T]]() extends AsmPhase[T] {
	var minLen = 0
	def runImpl(data: AssemblyData[T]): Unit = {
		val cs = makeContigSet(data.config.minKmers, data.graph, data.kmers)
		data.contigSet = cs
	}
  
	/**
	 * Produce a contig set.
	 * @param minLen minimum contig length
	 * @param graph The graph to read contigs from
	 * @param k The kmer size
	 * @param kmers The kmer set that the graph has been built from
	 */
	def makeContigSet(minLen: Int, graph: Graph[T], kmers: KmerSet[T]): ContigSet[T] = {
	  	this.minLen = minLen
		val contigSet = new ContigSet[T](kmers)
		val paths = findPaths(graph, kmers.k)
		for ((p, r1, r2) <- paths) {
		  contigSet.addContig(new Contig[T](contigSet, p, Some(r1), Some(r2)))
		}
	  	contigSet
	}
	
	/**
	 * Find contigs in the graph.
	 * Returns an iterable of the contig itself, and the reasons for stopping in the forward and the backward direction.
	 */
	def findPaths(g: Graph[T], k: Int): Iterable[(Seq[T],String,String)] = {
		
		var seen = new HashSet[T]()
		var res = List[Tuple3[Seq[T], String, String]]()
		
		var count = 0
		
		for (n <- g.nodes; if !seen.contains(n); if !n.ignore) {
			
			val fwd = findSeqFwd(n, seen, new ArrayBuffer[T]())
//			seen ++= fwd.map {_.id}
			seen -= n
			val back = findSeqBack(n, seen, new ArrayBuffer[T]())
//			seen ++= back.map {_.id}
			
			val fin = back._1.drop(1).reverse ++ fwd._1
			
			count += 1
			if (count % 100 == 0) {
				print (".")
			}
			if (count % 1000 == 0) {
				println("")
			}

			if (fin.size >= minLen) {
				//FindPathsPhase.printPath(fin, k)
				res ::= (fin, fwd._2, back._2)
			}

		}
		

		/**
		 * Find a path by going forward.
		 * @param start: node to start from. seen: Set of node ids that we have processed so far (in any contig)		 
		 * @param worklist: accumulator param. for the result.
		 * @returns The path and the reason for stopping.
		 */ 
		@tailrec def findSeqFwd(start: T, seen: HashSet[T], worklist:ArrayBuffer[T]): (ArrayBuffer[T], String) = {
		
			
			if (worklist.size % 1000 == 0 && worklist.size > 999) {
				print (">")
			}
			if (worklist.size % 10000 == 0 && worklist.size > 1) {
				println("")
			}
			
			if (seen.contains(start)) {
				//loop
				//TODO what if this is empty?
				(worklist, "Already seen node: " + start)
			} else {			
			
				worklist += start
				seen += start

				val validEdgesFrom = g.edgesFrom(start).filter { ! _.tip } 
				if (validEdgesFrom.isEmpty) {
					(worklist, "No edges")
				} else if (validEdgesFrom.size > 1) {
                    // now filter out low-coverage edges
                    val highCovEdges = validEdgesFrom.filter { ! _.noise }
                    if (highCovEdges.isEmpty)
                        (worklist, "No valid edges")
                    else if (highCovEdges.size > 1)
                        (worklist, "Multiple paths: " + validEdgesFrom.map(x => x + " " + x.coverage.toString))
                    else
                        findSeqFwd(highCovEdges.head, seen, worklist)
				} else {
					//exactly 1
					findSeqFwd(validEdgesFrom.head, seen, worklist)				
				}
			}
		}
		
		/**
		 * Find a path by going backward.
		 * @param start: node to start from. seen: Set of node ids that we have processed so far (in any contig)		 
		 * @param worklist: accumulator param. for the result.
		 * @returns The path and the reason for stopping.
		 */ 
		@tailrec def findSeqBack(start: T, seen: HashSet[T], worklist: ArrayBuffer[T]): (ArrayBuffer[T], String) = {
			assert(start != -1)			
			
			if (worklist.size % 1000 == 0 && worklist.size > 999) {
				print ("<")
			}
			if (worklist.size % 10000 == 0 && worklist.size > 1) {
				println("")
			}
						
			if (seen.contains(start)) {
				//loop
				(worklist, "Already seen node: " + start)
			} else {
			
				worklist += start
				seen += start

				val validEdgesTo = g.edgesTo(start).filter {! _.tip } 
				if (validEdgesTo.isEmpty) {
					(worklist, "No valid edges")
				} else if (validEdgesTo.size > 1) {
                    //now filter out low-coverage edges
                    val highCovEdges = validEdgesTo.filter { ! _.noise }
                    if (highCovEdges.isEmpty)
                        (worklist, "No valid edges")
                    else if (highCovEdges.size > 1)
                        (worklist, "Multiple paths: " + validEdgesTo.map(x => x + " " +  x.coverage.toString))
                    else
                        findSeqBack(highCovEdges.head, seen, worklist)
				} else {
					//exactly 1
					findSeqBack(validEdgesTo.head, seen, worklist)
				}
			}
		}		
		res						
	}
}
