package miniasm

import java.io._
import scala.actors.Actor
import miniasm.genome._
import miniasm.genome.bpbuffer._
import scala.collection.mutable.ArrayBuffer
import scala.io.BufferedSource

object ScanReadsPhase {
  /**
   * Convenience function for interactive use
   */
  def apply[T <: Kmer[T]](data: AssemblyData[T]) = (new ScanReadsPhase[T]).run(data)
}

/**
 * Phase to scan reads from an input file.
 * input: data.k
 * output: data.reads, data.kmers 
 *
 */
class ScanReadsPhase[T <: Kmer[T]] extends AsmPhase[T] with KmerInputProcessor[T] {
	  
	def runImpl(data: AssemblyData[T]) {
		var skip = 0
		var offset = 0
		val file = data.config.input
		
		if (file.endsWith(".fq")) {
			println("Fastq format")
			skip = 4
			offset = 1
		} else if (file.endsWith(".fa")) {
			println("Fasta format")
			skip = 2
			offset = 1
		} else {
			throw new Exception("fasta or fastq file format expected!")
		}
		initialise(data.k, data)
		
		getReads(file, skip, offset, data.config.mates, data.config.trackReads)
		val setAndReads = getProcessedReads()
		data.kmers = setAndReads._1
		data.reads = setAndReads._2
		println(data.reads.size + " reads including rev. complements")		
	}
	
	private def getReads(file: String, skip: Int, offset: Int, 
			mateFile: Option[String], trackReads: Boolean) {
	    val source = new BufferedSource(new FileInputStream(file))
	    val mateSource = mateFile.map((x: String) => {new BufferedSource(new FileInputStream(x)) })
	
		val relevantMainLines = source.getLines.grouped(skip).map(_(offset))
		val relevantMates = mateSource.map(_.getLines.grouped(skip).map(_(offset)))
		
		val linePairs = relevantMates match {
		  case None => {
		    handleReads(relevantMainLines, trackReads)		    		    
		  }
		  case Some(mateLines) => {		    
			  handleReadsAndMates(relevantMainLines, mateLines, trackReads)		    		   
		  }
		}				
	}
}

trait KmerInputProcessor[T <: Kmer[T]] extends DNAHelpers {
 	private var kmerSet: KmerSet[T] = null	
	private var reads = new ArrayBuffer[Read[T]](100000)
	
	private def isGoodRead(s: String) = (!s.contains("N") && s.size >= kmerSet.k)
	
	def initialise(k: Short, data: AssemblyData[T]) {
 	  kmerSet = data.factory.makeKmerSet(k)
 	}
	
	private def handleRead(read: String, trackReads: Boolean) {
		if (isGoodRead(read)) {			
			val r1 = kmerSet.processRead(read, trackReads)
			val r2 = kmerSet.processRead(reverseComplement(read), trackReads)
			reads += r1; reads += r2
		}
	}
	
	private def handleReadAndMate(read: String, mate: String, trackReads: Boolean) {	  
	  if (mate == null) {
	    handleRead(read, trackReads)
	    return
	  }
	  val g1 = isGoodRead(read)
	  val g2 = isGoodRead(mate)
	  if (!g2) {		
		  handleRead(read, trackReads)
	  } else if (!g1) {
		  handleRead(mate, trackReads)
	  } else if (g1 && g2) {
		  val r1 = kmerSet.processRead(read, trackReads)
		  val r2 = kmerSet.processRead(reverseComplement(read), trackReads)

		  val m1 = kmerSet.processRead(mate, trackReads)
		  val m2 = kmerSet.processRead(reverseComplement(mate), trackReads)
		  r1.mates ::= m1; r2.mates ::= m2
		  m1.mates ::= r1; m2.mates ::= r2
		  reads += r1; reads += r2; reads += m1; reads += m2
	  }		
	}
	
	protected def handleReads(reads: Iterator[String], trackReads: Boolean) {
		while(reads.hasNext) {
			handleRead(reads.next, trackReads)
		}	  
	}
	
	protected def handleReadsAndMates(reads: Iterator[String], mates: Iterator[String], trackReads:Boolean) {	  
	  while(reads.hasNext) {
		val read = reads.next
		val mate = mates.next
	    handleReadAndMate(read, mate, trackReads)
	  }
	}
	
	protected def getProcessedReads(): (KmerSet[T], Iterable[Read[T]]) = {
	  (kmerSet, reads)
	}
	
	def getKmerSet = kmerSet
	
	def dropData() = {
	   kmerSet = null
    }
	
}
 