/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package miniasm.util
import miniasm.genome.Kmer
import friedrich.graph._

/**
 * A graph node that can store two edges, "next" and "previous".
 */
trait MigratingGraphNode[N] {  	
	var nextNode: MigratingGraphNode[N] = null
	var prevNode: MigratingGraphNode[N] = null	
}

/**
 * Optimised version of the adj. list graph for the case with relatively
 * few edges. When a node has only two edges (one in and one out),
 *  they are stored in the node itself.
 * Note that counting all edges is an expensive operation.
 * 
 * TODO: override hasEdge
 */
trait MigratingGraph[N <: MigratingGraphNode[N] with Kmer[N]] extends AdjListGraph[N] {

	private def hasMore(node: N) = node.migrated 

	/**
	 * This is a statistics variable that counts how many nodes have been migrated
	 * from node-local storage to in-graph storage of the adjacency list.
	 */
	var migrations = 0
	
	
	//Speed up a common case when we construct the graph
	//TODO: is this method redundant? Leave it to the superclass method?
	final override def addEdge(from: N, to: N): Unit = {
	  if (!hasMore(from)) {
	    if (from.nextNode != to) {	      	      
	      implAddEdge(from, to)	      
	    } 
	  } else {
	    super.addEdge(from, to)
	  }		
	}

	//The ClassManifest is needed in order to permit creation of arrays
	implicit def nodeManifest: ClassManifest[N]
	
	override def edgesFrom(from: N): Iterable[N] = {
		if (!hasMore(from)) {
			if (from.nextNode == null) {			  
				Seq.empty
			} else {
				Array(from.nextNode.asInstanceOf[N])
			}
		} else {	
			adjList.getOrElse(from, Set())
		}

	}
	
	override def edgesTo(to: N): Iterable[N] = {
		if (!hasMore(to)) {
			if (to.prevNode == null) {
				Seq.empty
			} else {
				Array(to.prevNode.asInstanceOf[N])
			}
		} else {
			revAdjList.getOrElse(to, Set())
		}
	}
	
	final override def implAddEdge(from: N, to: N) {
		
		//migrate from node-local storage (prevNode and nextNode)
		//to storing the adjacency list in this object
		def migrate(n: N) {
//			println("Migrated")
			
			//must now guarantee that it is in both lists
//			adjList.add(n, Set())
//			revAdjList.add(n, Set())
			
			if (n.nextNode != null) {
				adjList.addBinding(n, n.nextNode.asInstanceOf[N])				
			}
			if (n.prevNode != null) {
				revAdjList.addBinding(n, n.prevNode.asInstanceOf[N])		
			}
			migrations += 1
			n.migrated = true
		}
		
		if (!hasMore(from)) {
			if (from.nextNode == null) {
				from.nextNode = to
			} else {
				migrate(from)	
				adjList.addBinding(from, to)							
			}		
		} else {
			adjList.addBinding(from, to)			
		}
		
		if (!hasMore(to)) {
			if (to.prevNode == null) {
				to.prevNode = from
			} else {
				migrate(to)
				revAdjList.addBinding(to, from)				
			}
		} else {
			revAdjList.addBinding(to, from)		
		}		

	}
	
}
