/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm.util

import miniasm.genome._
import scala.collection.immutable.List
import friedrich.graph._

/**
 * A graph that performs path analysis.
 */
trait PathGraph[N <: Kmer[N]] extends Graph[N] {
	
    def k: Int

    /**
     * The first node shared by the two paths.
     */
    def pathOverlap(path1: Iterable[N], path2: Iterable[N]): Option[N] = {      
        if ((path1 == Nil) || (path2 == Nil)) {
            return None
        }
        else if (path2.exists(_ == path1.head)) {
            return Some(path1.head)
        }
        else {
            pathOverlap(path1.drop(1), path2)
        }
    }

    /**
     * The maximum unambiguous forward path from node.
     */
    def followPathForward(from: N): Iterable[N] = {
        var steps = List[N]()
        var current = from
        // need to make sure the path is unambiguous in BOTH directions
        while ((edgesFrom(current).size == 1) && (edgesTo(current).size == 1)) {
            steps = current :: steps
            current = edgesFrom(current).head
        }
        steps
    }

    def expandNode(gkmer: N, n: Int): Seq[N] = expandNodeHelper(gkmer, n, gkmer)

    private def expandNodeHelper(gkmer: N, n: Int, nogo: N): Seq[N] = {
        if ((n == 0) || (edges(gkmer).toList.filterNot(_ == nogo).size == 0))
            return List[N](gkmer)
        List[N](gkmer) ++ edges(gkmer).toList.filterNot(_ == nogo).map(expandNodeHelper(_, n - 1, gkmer)).reduceLeft(_ ++ _)
    }

    /**
     * The maximum unambiguous backward path from the given node.
     */
    def followPathBack(from: N): Iterable[N] = {
        var steps = List[N]()
        var current = from
        // need to make sure the path is unambiguous in BOTH directions
        while ((edgesTo(current).size == 1) && (edgesFrom(current).size == 1)) {
            steps = current :: steps
            current = edgesTo(current).head
        }
        steps
    }

    def peekAhead(from: N): Iterable[N] = {
        stepsFrom(from, k + 1)
    }

    /**
     * A list of the next n nodes from the given node.
     * If unambiguous steps cannot be taken, an empty list is returned.
     */
    def stepsFrom(from: N, n: Int): Iterable[N] = {     
        var steps = List[N]()
        var current = from
        while ((edgesFrom(current).size == 1) && (steps.size < n)) {
            steps = current :: steps
            current = edgesFrom(current).head
        }
        //println(steps.size + "," + n)
        if (steps.size == n)
            steps
        else
            List()
    }

    /**
     * A list of the previous n nodes from the given node.
     * If unambiguous steps cannot be taken, an empty list is returned.
     */
    def backstepsFrom(from: N, n: Int): Iterable[N] = {
        // 
        var steps = List[N]()
        var current = from
        while ((edgesTo(current).size == 1) && (steps.size < n)) {
            steps = current :: steps
            current = edgesTo(current).head
        }
        if (steps.size == n)
            steps
        else
            List()
    }

    /**
     * A list of nodes that have multiple forward edges.
     */
    def findForwardJunctions(): Iterable[N] = {     
        var jnodes = List[N]()
        jnodes ++= nodes.filter(edgesFrom(_).filter(!_.ignore).size > 1)
        jnodes
    }

    /***
     * A list of nodes that have multiple backward edges.
     */
    def findBackwardJunctions(): Iterable[N] = {
        var jnodes = List[N]()
        jnodes ++= nodes.filter(edgesTo(_).filter(!_.ignore).size > 1)
        jnodes
    }

    def ignorePath(path: Iterable[N]) = { path.foreach { _.noise = true } }

    def unignorePath(path: Iterable[N]) = { path.foreach { _.noise = true } }

    def ignorePaths(paths: Iterable[Iterable[N]]) = { paths.foreach { ignorePath(_) } }

    def computePathCov(path: Iterable[N]): Float = { path.map (_.coverage).reduceLeft(_+_) / path.size }

    /**
     * Finds bubbles
     */
    def findBubbles() = {
        var count = 1
        var gcount = 0
        var found = false
        var paths = List[List[N]]()
        val junctions = findForwardJunctions()
        println("total of " + junctions.size + " branches to inspect")
        for (split <- junctions) {
            //println("branch node: " + split)
            paths = edgesFrom(split).map(_ :: Nil).toList
            while ((!found) && (count < k * 3)) {
                paths = transform(paths)
                if (haveIntersection(paths)) {
                    paths = paths.map(_.reverse)
                    ignorePaths(paths)
                    unignorePath(paths.sortWith((x,y) => computePathCov(x) < computePathCov(y)).head)
                    println("reduced bubble")
                    found = true
                }
                count += 1
            }
            found = false
            count = 1
            gcount += 1
            if (gcount % 100 == 0) { println(gcount + "/" + junctions.size + " done") }
        }

        def haveIntersection(paths: List[List[N]]) = { paths.map(_.toSet).reduceLeft(_.intersect(_)).size > 0 }

        def transform(list: List[List[N]]): List[List[N]] = {
            list match {                
                case ((m::ms)::ls) => {
                    if (edgesFrom(m).size > 0) {
                        edgesFrom(m).toList.map(_ :: m :: ms) ++ transform(ls)
                    }
                    else {
                        (m :: ms) :: transform(ls)
                    }
                }
                case Nil::Nil => Nil
                case Nil => Nil
            }
        }
    }
}

class SetBackedPathGraph[N <: Kmer[N]](val k: Int) extends SetBackedGraph[N] with PathGraph[N] with AdjListGraph[N]


