/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm.genome

/**
 * A set of contigs that resulted from an assembly operation.
 * @param kmerSet The kmer set that contains the raw data for this assembly.
 * 
 * TODO most methods here are not yet implemented.
 */
class ContigSet[T <: Kmer[_]](val kmerSet: KmerSet[T], 
    var contigs: Set[Contig[T]] = Set[Contig[T]]()) {
  
  def filter(f: Contig[T] => Boolean): ContigSet[T] = new ContigSet(kmerSet, contigs.filter(f))
  
  /**
   * Return the set of contigs that are supported by both contig sets.
   * Returned contigs may be trimmed as necessary.
   */
  def intersect(o: ContigSet[T]): ContigSet[T] = filter(o.contains(_))
  
  /**
   * Return the set of contigs that are supported by both contig sets.
   * Returned contigs may be extended as necessary.
   * 
   */
  def union(o: ContigSet[T]): ContigSet[T] = new ContigSet(kmerSet, contigs ++ o.contigs)  
  
  /**
   * Return the set of contigs that exist in either contig set,
   * but not in both.
   */
  def differences(o: ContigSet[T]): ContigSet[T] = filter(!o.contains(_)) union o.filter(!contains(_))
    
  /**
   * Does this contig set contain a supercontig of the given contig?
   * Also returns true if the given contig is contained exactly as specified.
   */
  def supports(c: Contig[T]): Boolean = false
  
  /**
   * Return all the supercontigs of the given contig.
   */
  def supporting(c: Contig[T]): ContigSet[T] = new ContigSet(kmerSet)
  
  /**
   * Does this contig set contain the given contig exactly as specified?
   */
  def contains(c: Contig[T]): Boolean = {
    contigs.contains(c)
  }
  
  def addContig(c: Contig[T]) {
    contigs += c
  }
  
  def addContig(kmers: Seq[T]): Contig[T] = {
    val ctg = new Contig[T](this, kmers)
    addContig(ctg)
    ctg
  }
  
  /**
   * The number of contigs contained in this contig set.
   */
  def size: Int = contigs.size
}