/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package miniasm.genome
import java.nio.CharBuffer

import miniasm.util._
import friedrich.graph._

/**
 * A kmer representation. The size is controlled by the overall parameter 'k'.
 * Kmer instances should normally be created from the KmerSet.
 * 
 */

trait Kmer[T <: Kmer[T]] extends DNAHelpers with MigratingGraphNode[T] with BPSequence {
	
  implicit private def intToByte(i: Int) = i.toByte
  
  /**
   * The coverage of this KMer, i.e., how many times it was seen in the reads.
   */
	var coverage: Int
	
    var status: Byte = 0x0 //used for keeping track of tips, etc
    /*  meaning of bits:
            bit 1 - tip
            bit 2 - noise
            bit 3 - marked for display in graph output
            bit 4 - migration status in the graph
    */

    /**
     * Get the "tip" flag, which indicates whether this KMer is a tip. 
     */
    def tip = (status & 0x1) == 0x1
    
    /**
     * Set the "tip" flag, which indicates whether this KMer is a tip. 
     */
    def tip_=(v: Boolean) { if (v) { status = (status | 0x1) }
                            else { status = (status & ~0x1) } }

	/**
     * Get the "noise" flag, which indicates whether this KMer is noise. 
     */
    def noise = (status & 0x2) == 0x2
    
    /**
     * Set the "noise" flag, which indicates whether this KMer is noise. 
     */
    def noise_=(v: Boolean) { if (v) { status = (status | 0x2) }
                              else { status = (status & ~0x2) } }

    def marked = (status & 0x4) == 0x4
    def marked_=(v: Boolean) { if (v) { status = (status | 0x4) }
                              else { status = (status & ~0x4) } }
        
    /**
     * For graph optimisation purposes
     */
    def migrated = (status & 0x8) == 0x8
    def migrated_=(v: Boolean) { if (v) { status = (status | 0x8) }
                              else { status = (status & ~0x8) } }
    
    /**
     * For equality optimisation/ kmer set purposes
     */
    def canonical = (status & 0x10) == 0x10
    def canonical_=(v: Boolean) { if (v) { status = (status | 0x10) }
                              else { status = (status & ~0x10) } }
    
    
    /**
     * Should this KMer be ignored? 
     */
    def ignore = (status & 0x3) > 0

    /**
     * Get the KmerSet that this KMer belongs to. 
     */
    def set: KmerSet[T]

    /**
     * Number of BPs in this KMer.
     */
    def k: Short = set.k

    /**
     * The contig number is stored here when needed.
     */
    var contig: Short = -1 	
	
    /**
     * Obtain the corresponding reverse complement KMer in the KmerSet. 
     */
	def reverseComplementKmer = set.findOrCreate(reverseComplement(this.toString))
	
	/**
	 * A continuation is a kmer that could come after this one in a contig.
	 * For example, continuations of ACTG are CTGA, CTGC, CTGT, CTGG. 
	 * This method returns the continuations in the same KmerSet that this KMer belongs to.
	 */
	def continuationsInSet: Seq[T]
    
    /**
     * A "continuesFrom" is the opposite of a continuation: a KMer that could come
     * before this one in a contig. 
     * This method returns the "continuesFroms" in the same KmerSet that this KMer belongs to.
     */
	def continuesFromInSet: Seq[T]
    
    override def toString(): String = toBPString
	
}
