/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm.genome
import miniasm.util._
import scala.collection.mutable

/**
 * A Kmer creation and management facility. When KMers are initially obtained, in the form of 
 * reads, they are registered here, which affects their coverage count.
 * They can also be obtained without affecting the coverage count.
 * Logically, each KmerSet corresponds to one sequencing experiment or one data file.
 */
trait KmerSet[T <: Kmer[_]] {
  
	implicit def toShort(i: Int) = i.asInstanceOf[Short]
  
	/**
	 * All KMers in this KmerSet.
	 */
	def kmers: Iterable[T]
	
	/**
	 * The size of KMers in this KmerSet.
	 */
	def k: Short
	
	/**
	 * Find or create the given kmer increasing coverage count by one.
	 * To find a kmer without increasing coverage, use KmerSet.apply (set(x) if set is the KmerSet)
	 */
	def findOrCreate(kmer: String): T
	
	/**
	 * Find or create the given kmer increasing coverage count by one.
	 * To find a kmer without increasing coverage, use KmerSet.apply (set(x) if set is the KmerSet)
	 */
	def findOrCreate(kmer: T): T
	
	/**
	 * Find the kmer, or return null
	 */
	def apply(kmer: String): T
	
	/**
	 * The type signature says Kmer to keep it a bit simpler, but in practice, we only expect
	 * to support kmers of the same type as T.
	 * @param kmer
	 * @return
	 */
	def apply(kmer: T): T
	
	def contains(kmer: String):Boolean = (apply(kmer) != null)
	def contains(kmer: T):Boolean = (apply(kmer) != null)
	
	/**
	 * Register new reads and kmers, increasing coverage count.
	 */
	def processRead(read: String, trackReads: Boolean): Read[T]
	
	var kmerReads: mutable.MultiMap[T, Read[T]] = new mutable.HashMap[T, mutable.Set[Read[T]]] with mutable.MultiMap[T, Read[T]]
	
	/**
	 * Add a new read to our registry
	 */
	protected def registerRead(read: Read[T]) {	  
	  var i = 0
	  while (i < read.kmers.size) {
		  kmerReads.addBinding(read.kmers(i), read)
		  i += 1
	  }
	}
	
	/**
	 * What reads could the kmer potentially have come from?
	 */
	def reads(kmer: T): scala.collection.Set[Read[T]] = kmerReads(kmer) 
	
	/**
	 * What are all the possible mates of the reads that this kmer could have come from?
	 */
	def allMates(kmer: T): scala.collection.Set[Read[T]] = reads(kmer).flatMap(_.mates)	
	    
	/**
	 * Add all kmers from the other set. 
	 */
	def addAllFrom(other: KmerSet[T], trackReads: Boolean)
	
	private val startTime = System.currentTimeMillis
	
	/**
	 * Print performance statistics.
	 */
	def stats(): Unit = {
		val rate = (hits + misses)/(System.currentTimeMillis - startTime)		    		    
		  
		printf("%s Cache hits: %d misses: %d ratio: %.2f, rate: %.2f/ms\n", 
		    this.toString, hits, misses, 
		    hits/(hits + misses).toDouble, rate.toDouble)
	}
	
		
	final protected var hits = 0
	final protected var misses = 0
	
	final protected def addHit() { hits += 1 }
	final protected def addMiss() { misses += 1 }
	
	
}
