/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm.genome.bpbuffer
import java.util.Arrays
import miniasm.genome._




/**
 * The BPBuffer stores base pairs as "twobits", which are pairs of bits.
 * The actual data is stored in the Byte array "data". Offset is the offset into this array
 * where the BPs begin, where 0-3 are offsets in the first byte, 4-7 in the second, and so on.
 * Size is the number of BPs stored by this BPBuffer, starting from the offset.
 * The maximum storage capacity is achieved by offset = 0 and size = data.size * 4.
 * 
 * BPBuffers are immutable. Instead of modifying the underlying data, modifying operations
 * create new instances.
 * Whenever possible we try to avoid copying the underlying byte array, changing instead the offset
 * and size parameters. 
 * 
 * The internal storage layout of data is continuous left-to-right:
 * (... ... bp1 bp2) (bp3 bp4 bp5 bp6) (bp7 bp8 bp9 bp10) (... ...
 * for offset = 2, size = 10, where the parentheses delimit each individual byte.
 */
trait BPBuffer extends DNAHelpers with BPSequence {

	/**
	 * Number of BPs in this sequence.
	 */
	def size: Short
	
	/**
	 * Offset into the raw data array where this sequence starts (number of BPs).
	 * Multiply by two to get bit offset.
	 */
	def offset: Short
	
	/**
	 * The raw data that backs this BPBuffer. This can contain more data than the sequence
	 * itself, since it may be shared between multiple objects.
	 * The sequence represented by a BPBuffer object will be a subsequence of the
	 * one contained here.
	 */
	def data: Array[Byte]
	
	/**
	 * Obtain the (twobit) bp at a given offset.
	 */
	def apply(pos: Int): Byte	
	/**
	 * Obtain the ACTG bp at a given offset.
	 */
	def charAt(pos: Int): Char = BitRepresentation.twobitToChar(apply(pos))
	
	def bp(index: Int) = BitRepresentation.twobitToChar(apply(index)) 
	
	/**
	 * Construct a new bpbuffer by prepending a new bp before the existing ones.
	 */
	def prependTwobit(twobit: Byte): BPBuffer	
	
	/**
	 * Construct a new bpbuffer by appending a bp after the existing ones.
	 */
	def appendTwobit(twobit: Byte): BPBuffer
	
	
	/**
	 * Construct a new bpbuffer by removing bps from the start and returning the remainder.
	 */
	def drop(size: Int): BPBuffer	
	
	/**
	 * Construct a new bpbuffer by taking a number of bps from the start.
	 */
	def take(size: Int): BPBuffer	
	
	/**
	 * Construct a new bpbuffer from a subsequence of this one.
	 */
	def slice(from: Short, length: Short): BPBuffer = BPBuffer.wrap(this, from, length)

	/**
	 * Copy the raw data representing this sequence into a new integer array.
	 */
	def rebuildAsArray: Array[Int] 

	/**
	 * Print detailed information about this object, including internal
	 * storage and representation details.
	 */
	def printDetailed: Unit
	
	/**
	 * Compare with another BPBuffer for ordering.
	 */
	def compareWith(other: BPBuffer): Int
	
	def lessThan(other: BPBuffer): Boolean

}

/**
* A BPBuffer is a sequence of base pairs. It is a low level building block to support
* other classes in a storage-efficient way. 
*
* @author Johan
*/

object BPBuffer extends BitHelpers {

	implicit def toShort(i: Int) = i.asInstanceOf[Short]
	
	/**
	 * Creates a new bpbuffer from an ACTG string.
	 */
	final def wrap(str: String):BPBuffer = {
		new ForwardBPBuffer(stringToBytes(str), 0, str.size)
	}
	
	/**
	 * Creates a new bpbuffer from an ACTG string, with a given 0-based starting offset
	 * and size.
	 */
	final def wrap(str: String, offset: Short, size: Short):BPBuffer = {
		new ForwardBPBuffer(stringToBytes(str), offset, size)
	}

	/**
	 * Creates a new bpbuffer from an existing bpbuffer.
	 * This saves memory by reusing the objects from the first one.
	 * For instance, if the old one is ACTGACTG, and we do
	 * BPBuffer.wrap(x,2,4), then the result contains TGAC, but using the same backing
	 * buffer.
	 */
	
	final def wrap(buffer: BPBuffer, offset: Short, size: Short): BPBuffer = {
		if (size % 4 == 0) {
			assert(buffer.data.length >= size/4)
		} else {
			assert(buffer.data.length >= size / 4 - 1)	
		}
			
		if (buffer.isInstanceOf[BPBufferImpl]) {
			val bpb = buffer.asInstanceOf[BPBufferImpl]
			
			assert(size <= (bpb.size - offset - bpb.offset))
			assert(size > 0 && offset >= 0)
			
			new ForwardBPBuffer(bpb.data, bpb.offset + offset, size)
		} else {
			throw new Exception("Unexpected buffer implementation")
		}
	}
	
	/**
	 * The main implementation of the BPBuffer trait.
	 */
	trait BPBufferImpl extends BPBuffer {
		
	  /**
	   * The "backing buffer" which contains the raw data.
	   */
		val data: Array[Byte]
		
		/**
		 * Offset into the buffer to read from (in base pairs). There are 4 base pairs in one byte.
		 */
        val offset: Short
        /**
	 	 * Returns a "twobit" in the form of a single byte.
		 * Only the lowest two bits of the byte are valid. The others will be zeroed out.
		 * If x is a BPBuffer, then you can do e.g. x(0), x(1)... x(n-1) where n = size. 
		 * That syntax will invoke this method.
		 */
		
		final def apply(pos: Int): Byte = {
			assert(pos >= 0 && pos < size)
			
			val truePos: Int = offset + pos			
			val byte = truePos/4 
			val bval = data(byte)
			val localOffset = truePos % 4
			((bval >> (2 * (3-localOffset))) & 0x3)			
		}
		
		final def rebuildAsArray: Array[Int] = computeIntArray()
		
		
		/**
		 * Create a new int array that contains the data in this bpbuffer, starting from offset 0.
		 * This means that it's suitable for direct comparison and equality testing.
		 * (BPBuffers may otherwise contain equivalent data, but in different backing buffers,
		 * and starting from different offsets, so the backing buffers cannot easily be compared
		 * directly for equality testing)
		 */
		
		private def numIntsInArray =  if (size % 16 == 0) { size/16 } else { size/16 + 1} 
		 
		private def computeIntArray() = {
			var i = 0
			val ns = numIntsInArray
			val newData = new Array[Int](ns)
			while (i < ns) {
				newData(i) = computeIntArrayElement(i)
				i += 1
			}
			newData
		}
		
		/**
		 * One integer contains 4 bytes. This function computes a single such integer from the
		 * underlying backing buffer.
		 * 
		 * This function is the main bottleneck of kmer equality and hashing, 
		 * so it needs to be fast!
		 */		
		
		private def computeIntArrayElement(i: Int): Int = {
			val os = offset
		    val spo = size + os
			val shift = (os % 4) * 2
			val mask = (spo) % 4				
			var res = 0
			
			val finalByte = (if (mask == 0) {
				(spo / 4) - 1
			} else {
			  spo / 4
			})
	
			var pos = (os / 4) + i * 4
			var intshift = 32 + shift
				
			//adjust pos here, since some bytes may need to be used both for the final byte of one int
			//and for the first byte of the next
			pos -= 1
				
			var j = 0
			val dl = data.length
			while (j < 5 && pos + 1 <= finalByte) {					
				//process 1 byte for each iteration

				intshift -= 8
				pos += 1

				var src = data(pos)

				//mask redundant bits from the final byte
				if (pos == finalByte) {
					if (mask == 1) {
						src = src & (0xfc << 4)
					} else if (mask == 2) {
						src = src & (0xfc << 2)
					} else if (mask == 3) {
						src = src & 0xfc
					} 
					
				} 
				var shifted = (src & 0xff) 

					//adjust to position in the current int			
				if (intshift > 0) {
					shifted = (shifted << intshift)
				} else {
					shifted = (shifted >> (-intshift))
				}
			
				//The bits should be mutually exclusive
//				assert ((res & shifted) == 0)
				res |= shifted
//				print(binint(newData(i)) + " ")
				
				j += 1
			}	
			res
		}
			
		private def binbyte(byte: Byte) = byteToQuad(byte)
		private def binint(in: Int) = byteToQuad((in >> 24) & 0xff) + byteToQuad((in >> 16) & 0xff) + byteToQuad((in >> 8) & 0xff) + byteToQuad(in & 0xff) + "/" + java.lang.Integer.toString(in, 2)
		
		/**
		 * Equality checking of two BPBufferImpl classes. Do they represent the same
		 * sequence of BP's ?
		 * As with kmerHashCode, the major source of complexity here is the fact that
		 * equal kmers may have data starting at different offsets within an integer.
		 * The computeIntArrayElement function compensates for this.
		 */
		final def kmerEquals(obfr: BPBufferImpl): Boolean = {					
			if (size != obfr.size) {
				return false //try to fail fast
			} else {
				val n = numIntsInArray //must be same for both if they have same size
				var i = 0
				while (i < n) {
					if (computeIntArrayElement(i) != obfr.computeIntArrayElement(i)) {
						return false
					}
					i += 1
				}
				return true
			}					
		}
		
		/**
		 * Hash a kmer.
		 * This is a very frequent operation. Performance is crucial.
		 * The main difficulty in hashing kmers stems from the fact that equal kmers
		 * could have data that starts at different offsets, from the perspective of a single integer.
		 * The data could start at any of the 16 bps that a 32-bit int can contain.
		 * The computeIntArrayElement performs the necessary adjustments to make the data comparable.
		 * 
		 * Note that we get a slight speedup if we make this function a lazy val, instead of a
		 * def.
		 * On the other hand, we use more memory that way.
		 * 
		 */
		final def kmerHashCode: Int = {
						
			val n = numIntsInArray			
			var r = 0
			var i = 0
							
			r = computeIntArrayElement(0)
			i += 1
			
			while (i < n) {
				//This is a common way to compute hash codes, multiply by 41 and add the next, spreads out the bits
				r = (r * 41) + computeIntArrayElement(i)
				i += 1
			}			
			r
		}

		def toBPString: String = {
			bytesToString(data, offset, size)
		}
		
		override def toString(): String = toBPString

		//insert the twobit into orig, returning the or:ed byte. 
		//offset is a 0-3 value.
		private def writeTwobitAtOffset(orig: Byte, twobit: Byte, offset: Int): Byte = {				
			var insertPart = (twobit << (2 * (3 - offset)))
			(orig | insertPart).asInstanceOf[Byte]		
		}
		
		/**
		 * Append a new base pair at the end of this buffer, returning the result in a new
		 * buffer.
		 */
		final def appendTwobit(twobit: Byte): BPBuffer = {
			//remaining BP positions in the final byte
			val remainingSpace = (4 - (offset + size % 4)) % 4
			
			var newSize = 0
			var writeByte = 0
			
			if (remainingSpace > 0) {
				newSize = data.size
				writeByte = data.size - 1
			} else {
				newSize = data.size + 1
				writeByte = data.size
			}
			
			val newData = Arrays.copyOf(data, newSize)
			val writeOffset = (offset + size % 4) 
			
			newData(writeByte) = writeTwobitAtOffset(newData(writeByte), twobit, writeOffset)
			
			//TODO can shrink this array in some cases
			new ForwardBPBuffer(newData, offset, size + 1)
		}
		
		/**
		 * Prepend a new base pair at the beginning of this buffer, returning the
		 * result in a new buffer.
		 */
		final def prependTwobit(twobit: Byte): BPBuffer = {			
			val newSize = size + 1
			val newData: Array[Byte] = Array.fill((newSize - 1)/ 4 + 1)(0)
			newData(0) = (twobit << 6).asInstanceOf[Byte]
			
			for (i <- 0 until size) {				
				val writeByte = (i+1)/4
				val writeOffset = (i+1) % 4				
				
				newData(writeByte) = writeTwobitAtOffset(newData(writeByte), this(i),
						writeOffset)
			}
			new ForwardBPBuffer(newData, 0, size + 1)	
		}
		
		/**
		 * Remove a number of base pairs from the start of this buffer, returning the result
		 * in a new buffer.
		 */
		final def drop(amt: Int): BPBuffer = {
			assert(amt < size)
			new ForwardBPBuffer(data, offset + amt, size - amt)
		}
	
		/**
		 * Take a number of base pairs from the start of this buffer, returning the
		 * result in a new buffer.
		 */
		final def take(amt: Int): BPBuffer = {
			assert(amt < size)
			new ForwardBPBuffer(data, offset, amt)
		}
		
		def printDetailed = {
			println("BPBuffer size " + size + " offset " + offset + " " + getClass)
			print("Data ")
			data.foreach (x => { print(x + " ") })
			print(" Ints ")
			rebuildAsArray.foreach( x=> { print(x + " " + binint(x)) })
			println(" String " + toString)
			computeIntArray()
		}
	
		/**
		 * Compare with another BPBuffer for ordering.
		 */
		final def compareWith(other: BPBuffer): Int = {		  
		  var i = 0
		  while (i < size && i < other.size) {
		    if (apply(i) < other.apply(i)) {
		      return -1
		    } else if (apply(i) > other.apply(i)) {
		      return 1
		    } 
		    i += 1
		  }
		  //unequal length but equal bps so far
		  if (size > other.size) {
		    return 1
		  } else if (size < other.size) {
		    return -1
		  } else {
		    //perfectly equal
		    return 0
		  }
		}
		
		final def lessThan(other: BPBuffer): Boolean = compareWith(other) < 0
	}	
	
	private class ForwardBPBuffer(val data: Array[Byte], val offset: Short, val size: Short) extends BPBufferImpl {
		override def hashCode:Int = kmerHashCode
		override def equals(other: Any):Boolean = other match {
		  case bbi: BPBufferImpl => kmerEquals(bbi)
		  case _ => false
		}
	}
	
}

