/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */

package miniasm.genome.bpbuffer
import miniasm.genome._
import miniasm.util._
import scala.collection.mutable._

import scala.collection._
import java.nio.CharBuffer

/**
 * This class serves as a context within which kmers can be created.
 * It helps track properties such as coverage count.
 */

class BPKmerSet(val k: Short) extends KmerSet[BPKmer] {
	
	private def toKmer(key: BPBuffer) = new BPKmerImpl(this, key, 1)		
	 
	var kmers: mutable.HashSet[BPKmer] = HashSet[BPKmer]()
	
	/**
	 * Finds or creates the kmer, increasing the coverage count by one. 
	* This is the primary method for making new kmers.
	* To find the kmer without increasing coverage count, use KmerSet.apply(key).
	*/ 
	final def findOrCreate(km: String): BPKmer = findOrCreate(toKmer(BPBuffer.wrap(km)))
	
	/**
	 * Finds or creates the kmer, increasing the coverage count by one. 
	* This is the primary method for making new kmers.
	* To find the kmer without increasing coverage count, use KmerSet.apply(key).
	*/ 
	final def findOrCreate(km: BPBuffer): BPKmer = findOrCreate(toKmer(km)) 
	
	/**
	 * Finds or creates the kmer, increasing the coverage count by one. 
	* This is the primary method for making new kmers.
	* To find the kmer without increasing coverage count, use KmerSet.apply(key).
	*/ 
	final def findOrCreate(key: BPKmer): BPKmer = {		
		kmers.findEntry(key) match {
		  case Some(k) => {
		    k.coverage += 1		
			addHit()
		    k
		  }
		  case None => {
		    key.canonical = true//make this a permanent member
			kmers += key			
			addMiss()
		    key
		  }
		}		
	}

	final override def contains(kmer: BPKmer):Boolean = kmers.contains(kmer) //.asInstanceOf[BPKmer])
	final override def contains(kmer: String):Boolean = contains(toKmer(BPBuffer.wrap(kmer)))
	final def contains(kmer: BPBuffer):Boolean = kmers.contains(toKmer(kmer))
	
	/**
	 * Look up the kmer without changing the coverage count.
	 */
	final override def apply(key: BPKmer):BPKmer = kmers.findEntry(key).get //.asInstanceOf[BPKmer]).get
	
	/**
	 * Look up the kmer without changing the coverage count.
	 */
	final override def apply(key: String):BPKmer = apply(BPBuffer.wrap(key))
	
	/**
	 * Look up the kmer without changing the coverage count.
	 */
	final def apply(key: BPBuffer):BPKmer = kmers.findEntry(toKmer(key)).get
	
	private var readCount = 0
	
	/**
	 * make reads and kmers, incresing coverage count.
	 */
	final def processRead(read: String, trackReads: Boolean): Read[BPKmer] = {
	  //TODO here: bpbuffer.wrap will not do well if the size of the read is > 32k base pairs.
	  //We need to handle this scenario gracefully, possibly involving changes to the bpbuffer
	  //itself? - johan 28 May
		val rb = BPBuffer.wrap(read)
		val kmers = new Array[BPKmer](read.size - k + 1)
		var i = 0
		while (i <= read.size - k) {
		  val b = rb.slice(i, k)
	      kmers(i) = findOrCreate(b)
	      i += 1
		}
		
		val r = new Read(kmers)
		if (trackReads) {
			registerRead(r)
		}
		
		readCount += 1
		if (readCount % 10000 == 0 && readCount > 0) {
		  stats()
		}
		r
	}
	
	final def addAllFrom(other: KmerSet[BPKmer], trackReads: Boolean) {
	  println("Merging KMer sets")
	  for (ok <- other.kmers) {
	    if (kmers.contains(ok)) {
			kmers.findEntry(ok).get.coverage += ok.coverage					
		} else {
			kmers += ok	
			ok.canonical = true
			ok.set = this
		}
	  }
	  
	  if (trackReads) {
		  kmerReads ++= other.kmerReads
	  }
	}

}

//This private kmer implementation class is located here to prevent anybody else from creating a kmer.
private class BPKmerImpl(set: BPKmerSet, source: BPBuffer,
    var coverage: Int) extends BPKmer(set, source.data, source.offset)


