/**
 * Part of the Friedrich bioinformatics framework.
 * Copyright (C) Gabriel Keeble-Gagnere and Johan Nystrom-Persson 2010-2012.
 * Dual GPL/MIT license. Please see the files README and LICENSE for details.
 */
package miniasm.genome.bpbuffer
import scala.collection.immutable._

/**
 * Helper functions for working with a low level bit representation of nucleotide sequences.
 * (Companion object)
 */
object BitRepresentation {
	val A:Byte = 0x0
	val C:Byte = 0x1
	val G:Byte = 0x2
	val T:Byte = 0x3
	
	val twobits = List(A, C, T, G)
	
	val byteToQuad = new Array[String](256)
	var quadToByte:Map[String, Byte] = new HashMap[String, Byte]()	
		
	//precompute conversion table once
	for (i <- 0 to 255) {
		val b = i.asInstanceOf[Byte]
		val str = byteToQuadCompute(b)
		byteToQuad(b - Byte.MinValue) = str 
		quadToByte += ((str, b))
		
	}
	
	/**
	 * Convert a single BP from string representation to "twobit" representation.
	 */
	def charToTwobit(char: Char): Byte = {		
		char match {
			case 'A' => A
			case 'C' => C
			case 'T' => T
			case 'G' => G
		}
	}
	
	/**
	 * Convert a single BP from twobit representation to string representation.
	 */
	def twobitToChar(byte: Byte): Char = {
		byte match {
			case A => 'A'
			case C => 'C'
			case T => 'T'
			case G => 'G'
		}
	}

		
	/**
	 * Convert a single byte to the "ACTG" format (a 4 letter string)
	 */
	def byteToQuadCompute(byte: Byte): String = {
		var res = ""
		val chars = for (i <- 0 to 3) {
			val ptn = ((byte >> ((3 - i)*2)) & 0x3)
			val char = twobitToChar(ptn.asInstanceOf[Byte])
			res += char
		} 
		res		
	}
	
}

/**
 * Helper functions for working with a low level bit representation of nucleotide sequences.
 */
trait BitRepresentation {
	
	def A = BitRepresentation.A
	def C = BitRepresentation.C
	def T = BitRepresentation.T
	def G = BitRepresentation.G 
	def twobits = BitRepresentation.twobits 
	
	/**
	 * Convert a byte to a 4-character string.
	 */
	def byteToQuad(byte: Byte) = BitRepresentation.byteToQuad(byte - Byte.MinValue)
	
	/**
	 * Convert a 4-character string to a byte.
	 */
	def quadToByte(quad: String) = BitRepresentation.quadToByte(quad)
	

	implicit def toByte(int: Int) = int.asInstanceOf[Byte]
		
	//xor
	def complement(byte: Byte) = (byte ^ 0xff).asInstanceOf[Byte]
	
	/**
	 * Convert a 4-character string to a byte, starting from a given offset.
	 */
	def quadToByte(str: String, offset: Int): Byte = {
		
		if (str.size - offset < 4) {
			val rem = (4 - ((str.size - offset) % 4)) % 4
			val s = str + ("A" * rem)			
			quadToByte(s.substring(offset, offset+4))
		} else {
			quadToByte(str.substring(offset, offset+4))
		}
				
	}
	

	//note that some strings get the same bytes (if they end in some number of A:s)
	//This function is now hand tuned for speed, and to avoid the creation of
	//excessive objects.
	
	def stringToBytes(bps: String): Array[Byte] = {
		var i = 0
		val rsize = (bps.size - 1)/4
		var r = new Array[Byte](rsize + 1)
		while (i <= rsize) {
		  r(i) = quadToByte(bps, i*4)
		  i += 1
		}
		r
		
//		(0 to (bps.size - 1)/4).map( x => { quadToByte(bps, x*4) }).toArray	
	}
	
	//length: length of result string in chars
	def bytesToString(bytes: Array[Byte], offset: Int, size: Int): String = {
		var res = new StringBuilder()
		
		val startByte = offset / 4
		
		for (i <- startByte until bytes.size) {
			if (res.size < size) {
				if (i == startByte) {
					res.append(byteToQuad(bytes(i)).substring(offset % 4, 4))									
				} else {
					res.append(byteToQuad(bytes(i)))					
				}
			}
			
		}
		res.substring(0, size)		
	}
	
}